# @summary Install and configure a Demarche Simplifiees app
#
# Install and configure a Demarche Simplifiees app, aka DS.
#
# @example
#   include demarchessimplifiees
#
# @param config configuration handled by .env to configure DemarchesSimplifiees settings.
# @param vcsurl git repository where DS sources are hosted.
# @param vcsref git reference to be installed.
# @param install_dir where sources are installed.
# @param bind Configure which IP address to listen on. When `undef` Puma binds `127.0.0.1` and `::1`.
# @param sysaccount_name OS account name owner of files
# @param sysaccount_manage
#   If `true` a local OS user account is created.
#   when creation of users accounts is handle somewhere else, you can disable with `false`.
# @param postgresql_manage
#   If `true` the PostgreSQL server is installed and configured on the same node.
#   When PostgreSQL service is managed somewhere else, you can disable with `false`.
# @param postgresql_role Needs to be documented.
# @param ruby_manage If `true`, Ruby is installed.
# @param ruby_version The installed version of Ruby when `ruby_manage` is `true`.
# @param yarn_manage If `true`, Yarn is installed.
# @param nodejs_manage If `true`, NodeJS is installed.
# @param nodejs_version The installed version of NodeJS when `nodejs_manage` is `true`.
# @param hostauthorization
#   Use RubyOnRails ActionDispatch::HostAuthorization against DNS rebinding attacks.
#   Only one entry is accepted.
#
class demarchessimplifiees (
  Demarchessimplifiees::Config $config,
  Stdlib::HTTPSUrl $vcsurl = 'https://github.com/betagouv/demarches-simplifiees.fr.git',
  String[1] $vcsref = '2021-04-02-01',
  Stdlib::Absolutepath $install_dir = '/opt',
  Optional[Stdlib::IP::Address] $bind = undef,
  String[1] $sysaccount_name = 'demarches',
  Boolean $sysaccount_manage = true,
  Boolean $postgresql_manage = true,
  String $postgresql_role = 'ds',
  Boolean $ruby_manage = true,
  String[1] $ruby_version = '2.7.1',
  Boolean $yarn_manage = true,
  Boolean $nodejs_manage = true,
  String[1] $nodejs_version = '12.x',
  Optional[Stdlib::Fqdn] $hostauthorization = undef,
) {

  $_homedir = "/home/${sysaccount_name}"
  $_dsdir = "${install_dir}/demarches-simplifiees"
  $_bundle_requisits = 'libpq-dev'
  $_rails_db_schema_load = 'rails db:schema:load'
  $_rails_db_seed = 'rails db:seed'
  $_path = "/usr/sbin:/usr/bin:/sbin:/bin:${_dsdir}/bin:${_homedir}/rbenv/shims"

  $_vcsref_min = '2021-04-02-01'

  if versioncmp($vcsref, $_vcsref_min) < 0 {
    fail("Used demarchessimplifiees::vcsref ${vcsref} is lower than ${_vcsref_min}, minimal required.")
  }

  if $postgresql_manage {
    # Install PostgreSQL server
    # Grant membership to a role
    # Create a role, database and assign the correct permissions
    # Activate an extension 'unaccent' on database
    include postgresql::server

    postgresql::server::role { $postgresql_role:
      password_hash => postgresql::postgresql_password($config['db_username'], $config['db_password']),
      require       => Class[postgresql::server],
    }
    -> postgresql::server::db { $config['db_database']:
      user     => $config['db_username'],
      password => postgresql::postgresql_password($config['db_username'], $config['db_password']),
      owner    => $postgresql_role,
      require  => Class[postgresql::server],
    }

    -> postgresql::server::extension { "unaccent-${config['db_database']}":
      ensure    => present,
      database  => $config['db_database'],
      extension => 'unaccent',
      require   => Class[postgresql::server],
    }

    Postgresql::Server::Extension["unaccent-${config['db_database']}"]
    -> File["${_dsdir}/config/database.yml"]
    -> Exec[$_rails_db_schema_load]
    -> Exec[$_rails_db_seed]
  }

  if $nodejs_manage {
    class { 'nodejs':
      repo_url_suffix => $nodejs_version,
    }
  }

  if $yarn_manage {
    include yarn

    Class[yarn] -> Exec['yarn install']
  }

  if $ruby_manage {
    # Ensure permissions on /usr/bin/env
    class { 'rbenv':
      install_dir => "${_homedir}/rbenv",
      owner       => $sysaccount_name,
    }

    rbenv::plugin { 'rbenv/ruby-build':
    }

    -> rbenv::build { $ruby_version:
      global => true,
    }

    Rbenv::Build[$ruby_version]
    -> Exec['bundle install']
    -> Exec[$_rails_db_schema_load]
    -> Exec[$_rails_db_seed]
  }

  if $sysaccount_manage {
    user { $sysaccount_name:
      ensure     => present,
      managehome => true,
    }

    User[$sysaccount_name] -> Vcsrepo[$_dsdir]
    if $ruby_manage {
      User[$sysaccount_name] -> Class[rbenv] -> Exec['bundle install']
    }
  }

  vcsrepo { $_dsdir:
    ensure   => present,
    provider => git,
    source   => $vcsurl,
    revision => $vcsref,
    owner    => $sysaccount_name,
  }

  file { $_dsdir:
    ensure  => directory,
    mode    => '0750',
    require => Vcsrepo[$_dsdir],
  }

  file { "${_dsdir}/.env":
    ensure  => file,
    owner   => $sysaccount_name,
    mode    => '0600',
    content => epp('demarchessimplifiees/env.epp'),
    require => Vcsrepo[$_dsdir],
    notify  => Service[puma],
  }

  if $hostauthorization {
    file_line { "authorize ${hostauthorization}":
      ensure  => present,
      path    => "${_dsdir}/config/environments/development.rb",
      line    => "Rails.application.config.hosts << '${hostauthorization}'",
      require => Vcsrepo[$_dsdir],
      notify  => Service[puma],
    }
  } else {
    file_line { 'remove hostauthorization':
      ensure            => absent,
      path              => "${_dsdir}/config/environments/development.rb",
      match             => '^Rails.application.config.hosts << ',
      match_for_absence => true,
      require           => Vcsrepo[$_dsdir],
      notify            => Service[puma],
    }
  }

  exec { 'yarn install':
    user    => $sysaccount_name,
    path    => $_path,
    cwd     => $_dsdir,
    unless  => "test -d ${_dsdir}/node_modules",
    require => Vcsrepo[$_dsdir],
  }

  package { $_bundle_requisits :
    ensure => 'present'
  }

  exec { 'bundle install':
    command => 'bundle install --path vendor/bundle',
    user    => $sysaccount_name,
    path    => $_path,
    cwd     => $_dsdir,
    unless  => "test -d ${_dsdir}/vendor/bundle",
    timeout => '2000',
    require => [
      Vcsrepo[$_dsdir],
      Package[$_bundle_requisits],
    ],
  }

  # Remove the definition of database tps_test
  file { "${_dsdir}/config/database.yml":
    ensure  => file,
    content => file('demarchessimplifiees/database.yml'),
    mode    => '0600',
    owner   => $sysaccount_name,
    group   => 0,
    require => [
      Vcsrepo[$_dsdir],
    ],
  }

  # Loads a schema.rb file into the database
  exec { $_rails_db_schema_load:
    user        => $sysaccount_name,
    path        => $_path,
    cwd         => $_dsdir,
    require     => [
      File["${_dsdir}/config/database.yml"],
      Exec['bundle install'],
      File["${_dsdir}/.env"],
    ],
    subscribe   => Exec['bundle install'],
    refreshonly => true,
  }

  # Loads the seed data from db/seeds.rb
  exec { $_rails_db_seed:
    user        => $sysaccount_name,
    path        => $_path,
    cwd         => $_dsdir,
    require     => [
      Exec[$_rails_db_schema_load],
    ],
    subscribe   => Exec[$_rails_db_schema_load],
    refreshonly => true,
  }

  systemd::unit_file { 'puma.service':
    content => epp('demarchessimplifiees/puma.service.epp',
      {
        path            => $_path,
        sysaccount_name => $sysaccount_name,
        dsdir           => $_dsdir,
        bind            => $bind,
      }
    ),
    require => [
      Exec['bundle install'],
      Exec['yarn install'],
      File["${_dsdir}/config/database.yml"],
      File["${_dsdir}/.env"],
    ],
  }
  ~> service {'puma':
    ensure => 'running',
  }
}
